@@ -2,11 +2,14 @@
 // Options.cs
 //
 // Authors:
-//  Jonathan Pryor <jpryor@novell.com>
+//  Jonathan Pryor <jpryor@novell.com>, <Jonathan.Pryor@microsoft.com>
 //  Federico Di Gregorio <fog@initd.org>
+//  Rolf Bjarne Kvinge <rolf@xamarin.com>
 //
 // Copyright (C) 2008 Novell (http://www.novell.com)
 // Copyright (C) 2009 Federico Di Gregorio.
+// Copyright (C) 2012 Xamarin Inc (http://www.xamarin.com)
+// Copyright (C) 2017 Microsoft Corporation (http://www.microsoft.com)
 //
 // Permission is hereby granted, free of charge, to any person obtaining
 // a copy of this software and associated documentation files (the
@@ -29,8 +32,8 @@
 //
 
 // Compile With:
-//   gmcs -debug+ -r:System.Core Options.cs -o:NDesk.Options.dll
-//   gmcs -debug+ -d:LINQ -r:System.Core Options.cs -o:NDesk.Options.dll
+//   mcs -debug+ -r:System.Core Options.cs -o:Mono.Options.dll -t:library
+//   mcs -debug+ -d:LINQ -r:System.Core Options.cs -o:Mono.Options.dll -t:library
 //
 // The LINQ version just changes the implementation of
 // OptionSet.Parse(IEnumerable<string>), and confers no semantic changes.
@@ -38,7 +41,7 @@
 //
 // A Getopt::Long-inspired option parsing library for C#.
 //
-// NDesk.Options.OptionSet is built upon a key/value table, where the
+// Mono.Options.OptionSet is built upon a key/value table, where the
 // key is a option format string and the value is a delegate that is 
 // invoked when the format string is matched.
 //
@@ -125,6 +128,33 @@
 //      p.Parse (new string[]{"-a-"});  // sets v == null
 //
 
+//
+// Mono.Options.CommandSet allows easily having separate commands and
+// associated command options, allowing creation of a *suite* along the
+// lines of **git**(1), **svn**(1), etc.
+//
+// CommandSet allows intermixing plain text strings for `--help` output,
+// Option values -- as supported by OptionSet -- and Command instances,
+// which have a name, optional help text, and an optional OptionSet.
+//
+//  var suite = new CommandSet ("suite-name") {
+//    // Use strings and option values, as with OptionSet
+//    "usage: suite-name COMMAND [OPTIONS]+",
+//    { "v:", "verbosity", (int? v) => Verbosity = v.HasValue ? v.Value : Verbosity+1 },
+//    // Commands may also be specified
+//    new Command ("command-name", "command help") {
+//      Options = new OptionSet {/*...*/},
+//      Run     = args => { /*...*/},
+//    },
+//    new MyCommandSubclass (),
+//  };
+//  return suite.Run (new string[]{...});
+//
+// CommandSet provides a `help` command, and forwards `help COMMAND`
+// to the registered Command instance by invoking Command.Invoke()
+// with `--help` as an option.
+//
+
 using System;
 using System.Collections;
 using System.Collections.Generic;
@@ -132,8 +162,12 @@ using System.Collections.ObjectModel;
 using System.ComponentModel;
 using System.Globalization;
 using System.IO;
+#if PCL
+using System.Reflection;
+#else
 using System.Runtime.Serialization;
 using System.Security.Permissions;
+#endif
 using System.Text;
 using System.Text.RegularExpressions;
 
@@ -145,6 +179,12 @@ using System.Linq;
 using NDesk.Options;
 #endif
 
+#if PCL
+using MessageLocalizerConverter = System.Func<string, string>;
+#else
+using MessageLocalizerConverter = System.Converter<string, string>;
+#endif
+
 #if NDESK_OPTIONS
 namespace NDesk.Options
 #else
@@ -193,7 +233,7 @@ namespace Mono.Options
 					if (char.IsWhiteSpace (c))
 						++start;
 					width = GetNextWidth (ewidths, width, ref hw);
-				} while (end < self.Length);
+				} while (start < self.Length);
 			}
 		}
 
@@ -372,13 +412,19 @@ namespace Mono.Options
 		OptionValueType type;
 		int count;
 		string[] separators;
+		bool hidden;
 
 		protected Option (string prototype, string description)
-			: this (prototype, description, 1)
+			: this (prototype, description, 1, false)
 		{
 		}
 
 		protected Option (string prototype, string description, int maxValueCount)
+			: this (prototype, description, maxValueCount, false)
+		{
+		}
+
+		protected Option (string prototype, string description, int maxValueCount, bool hidden)
 		{
 			if (prototype == null)
 				throw new ArgumentNullException ("prototype");
@@ -388,10 +434,19 @@ namespace Mono.Options
 				throw new ArgumentOutOfRangeException ("maxValueCount");
 
 			this.prototype   = prototype;
-			this.names       = prototype.Split ('|');
 			this.description = description;
 			this.count       = maxValueCount;
+			this.names       = (this is OptionSet.Category)
+				// append GetHashCode() so that "duplicate" categories have distinct
+				// names, e.g. adding multiple "" categories should be valid.
+				? new[]{prototype + this.GetHashCode ()}
+				: prototype.Split ('|');
+
+			if (this is OptionSet.Category || this is CommandOption)
+				return;
+
 			this.type        = ParsePrototype ();
+			this.hidden      = hidden;
 
 			if (this.count == 0 && type != OptionValueType.None)
 				throw new ArgumentException (
@@ -414,6 +469,7 @@ namespace Mono.Options
 		public string           Description     {get {return description;}}
 		public OptionValueType  OptionValueType {get {return type;}}
 		public int              MaxValueCount   {get {return count;}}
+		public bool             Hidden          {get {return hidden;}}
 
 		public string[] GetNames ()
 		{
@@ -430,15 +486,34 @@ namespace Mono.Options
 		protected static T Parse<T> (string value, OptionContext c)
 		{
 			Type tt = typeof (T);
-			bool nullable = tt.IsValueType && tt.IsGenericType && 
-				!tt.IsGenericTypeDefinition && 
-				tt.GetGenericTypeDefinition () == typeof (Nullable<>);
-			Type targetType = nullable ? tt.GetGenericArguments () [0] : typeof (T);
-			TypeConverter conv = TypeDescriptor.GetConverter (targetType);
+#if PCL
+			TypeInfo ti = tt.GetTypeInfo ();
+#else
+			Type ti = tt;
+#endif
+			bool nullable = 
+				ti.IsValueType && 
+				ti.IsGenericType && 
+				!ti.IsGenericTypeDefinition && 
+				ti.GetGenericTypeDefinition () == typeof (Nullable<>);
+#if PCL
+			Type targetType = nullable ? tt.GenericTypeArguments [0] : tt;
+#else
+			Type targetType = nullable ? tt.GetGenericArguments () [0] : tt;
+#endif
 			T t = default (T);
 			try {
-				if (value != null)
+				if (value != null) {
+#if PCL
+					if (targetType.GetTypeInfo ().IsEnum)
+						t = (T) Enum.Parse (targetType, value, true);
+					else
+						t = (T) Convert.ChangeType (value, targetType);
+#else
+					TypeConverter conv = TypeDescriptor.GetConverter (targetType);
 					t = (T) conv.ConvertFromString (value);
+#endif
+				}
 			}
 			catch (Exception e) {
 				throw new OptionException (
@@ -538,13 +613,110 @@ namespace Mono.Options
 
 		protected abstract void OnParseComplete (OptionContext c);
 
+		internal void InvokeOnParseComplete (OptionContext c)
+		{
+			OnParseComplete (c);
+		}
+
 		public override string ToString ()
 		{
 			return Prototype;
 		}
 	}
 
+	public abstract class ArgumentSource {
+
+		protected ArgumentSource ()
+		{
+		}
+
+		public abstract string[] GetNames ();
+		public abstract string Description { get; }
+		public abstract bool GetArguments (string value, out IEnumerable<string> replacement);
+
+#if !PCL || NETSTANDARD1_3
+		public static IEnumerable<string> GetArgumentsFromFile (string file)
+		{
+			return GetArguments (File.OpenText (file), true);
+		}
+#endif
+
+		public static IEnumerable<string> GetArguments (TextReader reader)
+		{
+			return GetArguments (reader, false);
+		}
+
+		// Cribbed from mcs/driver.cs:LoadArgs(string)
+		static IEnumerable<string> GetArguments (TextReader reader, bool close)
+		{
+			try {
+				StringBuilder arg = new StringBuilder ();
+
+				string line;
+				while ((line = reader.ReadLine ()) != null) {
+					int t = line.Length;
+
+					for (int i = 0; i < t; i++) {
+						char c = line [i];
+						
+						if (c == '"' || c == '\'') {
+							char end = c;
+							
+							for (i++; i < t; i++){
+								c = line [i];
+
+								if (c == end)
+									break;
+								arg.Append (c);
+							}
+						} else if (c == ' ') {
+							if (arg.Length > 0) {
+								yield return arg.ToString ();
+								arg.Length = 0;
+							}
+						} else
+							arg.Append (c);
+					}
+					if (arg.Length > 0) {
+						yield return arg.ToString ();
+						arg.Length = 0;
+					}
+				}
+			}
+			finally {
+				if (close)
+					reader.Dispose ();
+			}
+		}
+	}
+
+#if !PCL || NETSTANDARD1_3
+	public class ResponseFileSource : ArgumentSource {
+
+		public override string[] GetNames ()
+		{
+			return new string[]{"@file"};
+		}
+
+		public override string Description {
+			get {return "Read response file for more options.";}
+		}
+
+		public override bool GetArguments (string value, out IEnumerable<string> replacement)
+		{
+			if (string.IsNullOrEmpty (value) || !value.StartsWith ("@")) {
+				replacement = null;
+				return false;
+			}
+			replacement = ArgumentSource.GetArgumentsFromFile (value.Substring (1));
+			return true;
+		}
+	}
+#endif
+
+#if !PCL
 	[Serializable]
+#endif
 	public class OptionException : Exception {
 		private string option;
 
@@ -564,22 +736,28 @@ namespace Mono.Options
 			this.option = optionName;
 		}
 
+#if !PCL
 		protected OptionException (SerializationInfo info, StreamingContext context)
 			: base (info, context)
 		{
 			this.option = info.GetString ("OptionName");
 		}
+#endif
 
 		public string OptionName {
 			get {return this.option;}
 		}
 
+#if !PCL
+#pragma warning disable 618 // SecurityPermissionAttribute is obsolete
 		[SecurityPermission (SecurityAction.LinkDemand, SerializationFormatter = true)]
+#pragma warning restore 618
 		public override void GetObjectData (SerializationInfo info, StreamingContext context)
 		{
 			base.GetObjectData (info, context);
 			info.AddValue ("OptionName", option);
 		}
+#endif
 	}
 
 	public delegate void OptionAction<TKey, TValue> (TKey key, TValue value);
@@ -587,21 +765,36 @@ namespace Mono.Options
 	public class OptionSet : KeyedCollection<string, Option>
 	{
 		public OptionSet ()
-			: this (delegate (string f) {return f;})
+			: this (null)
 		{
 		}
 
-		public OptionSet (Converter<string, string> localizer)
+		public OptionSet (MessageLocalizerConverter localizer)
 		{
+			this.roSources = new ReadOnlyCollection<ArgumentSource> (sources);
 			this.localizer = localizer;
+			if (this.localizer == null) {
+				this.localizer = delegate (string f) {
+					return f;
+				};
+			}
 		}
 
-		Converter<string, string> localizer;
+		MessageLocalizerConverter localizer;
 
-		public Converter<string, string> MessageLocalizer {
+		public MessageLocalizerConverter MessageLocalizer {
 			get {return localizer;}
+			internal set {localizer = value;}
 		}
 
+		List<ArgumentSource> sources = new List<ArgumentSource> ();
+		ReadOnlyCollection<ArgumentSource> roSources;
+
+		public ReadOnlyCollection<ArgumentSource> ArgumentSources {
+			get {return roSources;}
+		}
+
+
 		protected override string GetKeyForItem (Option item)
 		{
 			if (item == null)
@@ -667,6 +860,31 @@ namespace Mono.Options
 			}
 		}
 
+		public OptionSet Add (string header)
+		{
+			if (header == null)
+				throw new ArgumentNullException ("header");
+			Add (new Category (header));
+			return this;
+		}
+
+		internal sealed class Category : Option {
+
+			// Prototype starts with '=' because this is an invalid prototype
+			// (see Option.ParsePrototype(), and thus it'll prevent Category
+			// instances from being accidentally used as normal options.
+			public Category (string description)
+				: base ("=:Category:= " + description, description)
+			{
+			}
+
+			protected override void OnParseComplete (OptionContext c)
+			{
+				throw new NotSupportedException ("Category.OnParseComplete should not be invoked.");
+			}
+		}
+
+
 		public new OptionSet Add (Option option)
 		{
 			base.Add (option);
@@ -677,7 +895,12 @@ namespace Mono.Options
 			Action<OptionValueCollection> action;
 
 			public ActionOption (string prototype, string description, int count, Action<OptionValueCollection> action)
-				: base (prototype, description, count)
+				: this (prototype, description, count, action, false)
+			{
+			}
+
+			public ActionOption (string prototype, string description, int count, Action<OptionValueCollection> action, bool hidden)
+				: base (prototype, description, count, hidden)
 			{
 				if (action == null)
 					throw new ArgumentNullException ("action");
@@ -696,11 +919,16 @@ namespace Mono.Options
 		}
 
 		public OptionSet Add (string prototype, string description, Action<string> action)
+		{
+			return Add (prototype, description, action, false);
+		}
+
+		public OptionSet Add (string prototype, string description, Action<string> action, bool hidden)
 		{
 			if (action == null)
 				throw new ArgumentNullException ("action");
 			Option p = new ActionOption (prototype, description, 1, 
-					delegate (OptionValueCollection v) { action (v [0]); });
+					delegate (OptionValueCollection v) { action (v [0]); }, hidden);
 			base.Add (p);
 			return this;
 		}
@@ -712,10 +940,14 @@ namespace Mono.Options
 
 		public OptionSet Add (string prototype, string description, OptionAction<string, string> action)
 		{
+			return Add (prototype, description, action, false);
+		}
+
+		public OptionSet Add (string prototype, string description, OptionAction<string, string> action, bool hidden)	{
 			if (action == null)
 				throw new ArgumentNullException ("action");
 			Option p = new ActionOption (prototype, description, 2, 
-					delegate (OptionValueCollection v) {action (v [0], v [1]);});
+					delegate (OptionValueCollection v) {action (v [0], v [1]);}, hidden);
 			base.Add (p);
 			return this;
 		}
@@ -776,48 +1008,30 @@ namespace Mono.Options
 			return Add (new ActionOption<TKey, TValue> (prototype, description, action));
 		}
 
+		public OptionSet Add (ArgumentSource source)
+		{
+			if (source == null)
+				throw new ArgumentNullException ("source");
+			sources.Add (source);
+			return this;
+		}
+
 		protected virtual OptionContext CreateOptionContext ()
 		{
 			return new OptionContext (this);
 		}
 
-#if LINQ
-		public List<string> Parse (IEnumerable<string> arguments)
-		{
-			bool process = true;
-			OptionContext c = CreateOptionContext ();
-			c.OptionIndex = -1;
-			var def = GetOptionForName ("<>");
-			var unprocessed = 
-				from argument in arguments
-				where ++c.OptionIndex >= 0 && (process || def != null)
-					? process
-						? argument == "--" 
-							? (process = false)
-							: !Parse (argument, c)
-								? def != null 
-									? Unprocessed (null, def, c, argument) 
-									: true
-								: false
-						: def != null 
-							? Unprocessed (null, def, c, argument)
-							: true
-					: true
-				select argument;
-			List<string> r = unprocessed.ToList ();
-			if (c.Option != null)
-				c.Option.Invoke (c);
-			return r;
-		}
-#else
 		public List<string> Parse (IEnumerable<string> arguments)
 		{
+			if (arguments == null)
+				throw new ArgumentNullException ("arguments");
 			OptionContext c = CreateOptionContext ();
 			c.OptionIndex = -1;
 			bool process = true;
 			List<string> unprocessed = new List<string> ();
 			Option def = Contains ("<>") ? this ["<>"] : null;
-			foreach (string argument in arguments) {
+			ArgumentEnumerator ae = new ArgumentEnumerator (arguments);
+			foreach (string argument in ae) {
 				++c.OptionIndex;
 				if (argument == "--") {
 					process = false;
@@ -827,6 +1041,8 @@ namespace Mono.Options
 					Unprocessed (unprocessed, def, c, argument);
 					continue;
 				}
+				if (AddSource (ae, argument))
+					continue;
 				if (!Parse (argument, c))
 					Unprocessed (unprocessed, def, c, argument);
 			}
@@ -834,7 +1050,50 @@ namespace Mono.Options
 				c.Option.Invoke (c);
 			return unprocessed;
 		}
-#endif
+
+		class ArgumentEnumerator : IEnumerable<string> {
+			List<IEnumerator<string>> sources = new List<IEnumerator<string>> ();
+
+			public ArgumentEnumerator (IEnumerable<string> arguments)
+			{
+				sources.Add (arguments.GetEnumerator ());
+			}
+
+			public void Add (IEnumerable<string> arguments)
+			{
+				sources.Add (arguments.GetEnumerator ());
+			}
+
+			public IEnumerator<string> GetEnumerator ()
+			{
+				do {
+					IEnumerator<string> c = sources [sources.Count-1];
+					if (c.MoveNext ())
+						yield return c.Current;
+					else {
+						c.Dispose ();
+						sources.RemoveAt (sources.Count-1);
+					}
+				} while (sources.Count > 0);
+			}
+
+			IEnumerator IEnumerable.GetEnumerator ()
+			{
+				return GetEnumerator ();
+			}
+		}
+
+		bool AddSource (ArgumentEnumerator ae, string argument)
+		{
+			foreach (ArgumentSource source in sources) {
+				IEnumerable<string> replacement;
+				if (!source.GetArguments (argument, out replacement))
+					continue;
+				ae.Add (replacement);
+				return true;
+			}
+			return false;
+		}
 
 		private static bool Unprocessed (ICollection<string> extra, Option def, OptionContext c, string argument)
 		{
@@ -956,7 +1215,7 @@ namespace Mono.Options
 					if (i == 0)
 						return false;
 					throw new OptionException (string.Format (localizer (
-									"Cannot bundle unregistered option '{0}'."), opt), opt);
+									"Cannot use unregistered option '{0}' in bundle '{1}'."), rn, f + n), null);
 				}
 				p = this [rn];
 				switch (p.OptionValueType) {
@@ -987,11 +1246,31 @@ namespace Mono.Options
 		}
 
 		private const int OptionWidth = 29;
+		private const int Description_FirstWidth  = 80 - OptionWidth;
+		private const int Description_RemWidth    = 80 - OptionWidth - 2;
+
+		static  readonly    string      CommandHelpIndentStart       = new string (' ', OptionWidth);
+		static  readonly    string      CommandHelpIndentRemaining   = new string (' ', OptionWidth + 2);
 
 		public void WriteOptionDescriptions (TextWriter o)
 		{
 			foreach (Option p in this) {
 				int written = 0;
+
+				if (p.Hidden)
+					continue;
+
+				Category c = p as Category;
+				if (c != null) {
+					WriteDescription (o, p.Description, "", 80, 80);
+					continue;
+				}
+				CommandOption co = p as CommandOption;
+				if (co != null) {
+					WriteCommandDescription (o, co.Command);
+					continue;
+				}
+
 				if (!WriteOptionPrototype (o, p, ref written))
 					continue;
 
@@ -1002,14 +1281,55 @@ namespace Mono.Options
 					o.Write (new string (' ', OptionWidth));
 				}
 
-				bool indent = false;
-				string prefix = new string (' ', OptionWidth+2);
-				foreach (string line in GetLines (localizer (GetDescription (p.Description)))) {
-					if (indent) 
-						o.Write (prefix);
-					o.WriteLine (line);
-					indent = true;
+				WriteDescription (o, p.Description, new string (' ', OptionWidth+2),
+						Description_FirstWidth, Description_RemWidth);
+			}
+
+			foreach (ArgumentSource s in sources) {
+				string[] names = s.GetNames ();
+				if (names == null || names.Length == 0)
+					continue;
+
+				int written = 0;
+
+				Write (o, ref written, "  ");
+				Write (o, ref written, names [0]);
+				for (int i = 1; i < names.Length; ++i) {
+					Write (o, ref written, ", ");
+					Write (o, ref written, names [i]);
 				}
+
+				if (written < OptionWidth)
+					o.Write (new string (' ', OptionWidth - written));
+				else {
+					o.WriteLine ();
+					o.Write (new string (' ', OptionWidth));
+				}
+
+				WriteDescription (o, s.Description, new string (' ', OptionWidth+2),
+						Description_FirstWidth, Description_RemWidth);
+			}
+		}
+
+		internal void WriteCommandDescription (TextWriter o, Command c)
+		{
+			var name = new string (' ', 8) + c.Name;
+			if (name.Length < OptionWidth - 1) {
+				WriteDescription (o, name + new string (' ', OptionWidth - name.Length) + c.Help, CommandHelpIndentRemaining, 80, Description_RemWidth);
+			} else {
+				WriteDescription (o, name, "", 80, 80);
+				WriteDescription (o, CommandHelpIndentStart + c.Help, CommandHelpIndentRemaining, 80, Description_RemWidth);
+			}
+		}
+
+		void WriteDescription (TextWriter o, string value, string prefix, int firstWidth, int remWidth)
+		{
+			bool indent = false;
+			foreach (string line in GetLines (localizer (GetDescription (value)), firstWidth, remWidth)) {
+				if (indent)
+					o.Write (prefix);
+				o.WriteLine (line);
+				indent = true;
 			}
 		}
 
@@ -1070,28 +1390,27 @@ namespace Mono.Options
 			o.Write (s);
 		}
 
-		private static string GetArgumentName (int index, int maxIndex, string description)
+		static string GetArgumentName (int index, int maxIndex, string description)
 		{
-			if (description == null)
-				return maxIndex == 1 ? "VALUE" : "VALUE" + (index + 1);
-			string[] nameStart;
-			if (maxIndex == 1)
-				nameStart = new string[]{"{0:", "{"};
-			else
-				nameStart = new string[]{"{" + index + ":"};
-			for (int i = 0; i < nameStart.Length; ++i) {
-				int start, j = 0;
-				do {
-					start = description.IndexOf (nameStart [i], j);
-				} while (start >= 0 && j != 0 ? description [j++ - 1] == '{' : false);
-				if (start == -1)
-					continue;
-				int end = description.IndexOf ("}", start);
-				if (end == -1)
-					continue;
-				return description.Substring (start + nameStart [i].Length, end - start - nameStart [i].Length);
+			var matches = Regex.Matches (description ?? "", @"(?<=(?<!\{)\{)[^{}]*(?=\}(?!\}))"); // ignore double braces 
+			string argName = "";
+			foreach (Match match in matches) {
+				var parts = match.Value.Split (':');
+				// for maxIndex=1 it can be {foo} or {0:foo}
+				if (maxIndex == 1) {
+					argName = parts[parts.Length - 1];
+				}
+				// look for {i:foo} if maxIndex > 1
+				if (maxIndex > 1 && parts.Length == 2 && 
+					parts[0] == index.ToString (CultureInfo.InvariantCulture)) {
+					argName = parts[1];
+				}
 			}
-			return maxIndex == 1 ? "VALUE" : "VALUE" + (index + 1);
+
+			if (string.IsNullOrEmpty (argName)) {
+				argName = maxIndex == 1 ? "VALUE" : "VALUE" + (index + 1);
+			}
+			return argName;
 		}
 
 		private static string GetDescription (string description)
@@ -1136,11 +1455,357 @@ namespace Mono.Options
 			return sb.ToString ();
 		}
 
-		private static IEnumerable<string> GetLines (string description)
+		private static IEnumerable<string> GetLines (string description, int firstWidth, int remWidth)
 		{
-			return StringCoda.WrappedLines (description, 
-					80 - OptionWidth, 
-					80 - OptionWidth - 2);
+			return StringCoda.WrappedLines (description, firstWidth, remWidth);
+		}
+	}
+
+	public class Command
+	{
+		public      string                              Name            {get;}
+		public      string                              Help            {get;}
+
+		public      OptionSet                           Options         {get; set;}
+		public      Action<IEnumerable<string>>         Run             {get; set;}
+
+		public      CommandSet                          CommandSet      {get; internal set;}
+
+		public Command (string name, string help = null)
+		{
+			if (string.IsNullOrEmpty (name))
+				throw new ArgumentNullException (nameof (name));
+
+			Name    = name;
+			Help    = help;
+		}
+
+		public virtual int Invoke (IEnumerable<string> arguments)
+		{
+			var rest    = Options?.Parse (arguments) ?? arguments;
+			Run?.Invoke (rest);
+			return 0;
+		}
+	}
+
+	class CommandOption : Option
+	{
+		public      Command             Command         {get;}
+
+		// Prototype starts with '=' because this is an invalid prototype
+		// (see Option.ParsePrototype(), and thus it'll prevent Category
+		// instances from being accidentally used as normal options.
+		public CommandOption (Command command, bool hidden = false)
+			: base ("=:Command:= " + command?.Name, command?.Name, maxValueCount: 0, hidden: hidden)
+		{
+			if (command == null)
+				throw new ArgumentNullException (nameof (command));
+			Command = command;
+		}
+
+		protected override void OnParseComplete (OptionContext c)
+		{
+			throw new NotSupportedException ("CommandOption.OnParseComplete should not be invoked.");
+		}
+	}
+
+	class HelpOption : Option
+	{
+		Option      option;
+		CommandSet  commands;
+
+		public HelpOption (CommandSet commands, Option d)
+			: base (d.Prototype, d.Description, d.MaxValueCount, d.Hidden)
+		{
+			this.commands   = commands;
+			this.option     = d;
+		}
+
+		protected override void OnParseComplete (OptionContext c)
+		{
+			commands.showHelp  = true;
+
+			option?.InvokeOnParseComplete (c);
+		}
+	}
+
+	class CommandOptionSet : OptionSet
+	{
+		CommandSet  commands;
+
+		public CommandOptionSet (CommandSet commands, MessageLocalizerConverter localizer)
+			: base (localizer)
+		{
+			this.commands = commands;
+		}
+
+		protected override void SetItem (int index, Option item)
+		{
+			if (ShouldWrapOption (item)) {
+				base.SetItem (index, new HelpOption (commands, item));
+				return;
+			}
+			base.SetItem (index, item);
+		}
+
+		bool ShouldWrapOption (Option item)
+		{
+			if (item == null)
+				return false;
+			var help = item as HelpOption;
+			if (help != null)
+				return false;
+			foreach (var n in item.Names) {
+				if (n == "help")
+					return true;
+			}
+			return false;
+		}
+
+		protected override void InsertItem (int index, Option item)
+		{
+			if (ShouldWrapOption (item)) {
+				base.InsertItem (index, new HelpOption (commands, item));
+				return;
+			}
+			base.InsertItem (index, item);
+		}
+	}
+
+	public class CommandSet : KeyedCollection<string, Command>
+	{
+		readonly    OptionSet       options;
+		readonly    TextWriter      outWriter;
+		readonly    TextWriter      errorWriter;
+		readonly    string          suite;
+
+		HelpCommand help;
+
+		internal    bool            showHelp;
+
+		internal    OptionSet       Options     => options;
+
+#if !PCL || NETSTANDARD1_3
+		public CommandSet(string suite, MessageLocalizerConverter localizer = null)
+			: this(suite, Console.Out, Console.Error, localizer)
+		{
+		}
+#endif
+		
+		public CommandSet (string suite, TextWriter output, TextWriter error, MessageLocalizerConverter localizer = null)
+		{
+			if (suite == null)
+				throw new ArgumentNullException (nameof (suite));
+			if (output == null)
+				throw new ArgumentNullException (nameof (output));
+			if (error == null)
+				throw new ArgumentNullException (nameof (error));
+
+			this.suite  = suite;
+			options     = new CommandOptionSet (this, localizer);
+			outWriter   = output;
+			errorWriter = error;
+		}
+
+		public  string                          Suite               => suite;
+		public  TextWriter                      Out                 => outWriter;
+		public  TextWriter                      Error               => errorWriter;
+		public  MessageLocalizerConverter       MessageLocalizer    => options.MessageLocalizer;
+
+		protected override string GetKeyForItem (Command item)
+		{
+			return item?.Name;
+		}
+
+		public new CommandSet Add (Command value)
+		{
+			if (value == null)
+				throw new ArgumentNullException (nameof (value));
+			AddCommand (value);
+			options.Add (new CommandOption (value));
+			return this;
+		}
+
+		void AddCommand (Command value)
+		{
+			if (value.CommandSet != null && value.CommandSet != this) {
+				throw new ArgumentException ("Command instances can only be added to a single CommandSet.", nameof (value));
+			}
+			value.CommandSet                = this;
+			if (value.Options != null) {
+				value.Options.MessageLocalizer  = options.MessageLocalizer;
+			}
+
+			base.Add (value);
+
+			help    = help ?? value as HelpCommand;
+		}
+
+		public CommandSet Add (string header)
+		{
+			options.Add (header);
+			return this;
+		}
+
+		public CommandSet Add (Option option)
+		{
+			options.Add (option);
+			return this;
+		}
+
+		public CommandSet Add (string prototype, Action<string> action)
+		{
+			options.Add (prototype, action);
+			return this;
+		}
+
+		public CommandSet Add (string prototype, string description, Action<string> action)
+		{
+			options.Add (prototype, description, action);
+			return this;
+		}
+
+		public CommandSet Add (string prototype, string description, Action<string> action, bool hidden)
+		{
+			options.Add (prototype, description, action, hidden);
+			return this;
+		}
+
+		public CommandSet Add (string prototype, OptionAction<string, string> action)
+		{
+			options.Add (prototype, action);
+			return this;
+		}
+
+		public CommandSet Add (string prototype, string description, OptionAction<string, string> action)
+		{
+			options.Add (prototype, description, action);
+			return this;
+		}
+
+		public CommandSet Add (string prototype, string description, OptionAction<string, string> action, bool hidden)
+		{
+			options.Add (prototype, description, action, hidden);
+			return this;
+		}
+
+		public CommandSet Add<T> (string prototype, Action<T> action)
+		{
+			options.Add (prototype, null, action);
+			return this;
+		}
+
+		public CommandSet Add<T> (string prototype, string description, Action<T> action)
+		{
+			options.Add (prototype, description, action);
+			return this;
+		}
+
+		public CommandSet Add<TKey, TValue> (string prototype, OptionAction<TKey, TValue> action)
+		{
+			options.Add (prototype, action);
+			return this;
+		}
+
+		public CommandSet Add<TKey, TValue> (string prototype, string description, OptionAction<TKey, TValue> action)
+		{
+			options.Add (prototype, description, action);
+			return this;
+		}
+
+		public CommandSet Add (ArgumentSource source)
+		{
+			options.Add (source);
+			return this;
+		}
+
+		public int Run (IEnumerable<string> arguments)
+		{
+			if (arguments == null)
+				throw new ArgumentNullException (nameof (arguments));
+
+			this.showHelp   = false;
+			if (help == null) {
+				help    = new HelpCommand ();
+				AddCommand (help);
+			}
+			Action<string>  setHelp     = v => showHelp = v != null;
+			if (!options.Contains ("help")) {
+				options.Add ("help", "", setHelp, hidden: true);
+			}
+			if (!options.Contains ("?")) {
+				options.Add ("?", "", setHelp, hidden: true);
+			}
+			var extra   = options.Parse (arguments);
+			if (extra.Count == 0) {
+				if (showHelp) {
+					return help.Invoke (extra);
+				}
+				Out.WriteLine (options.MessageLocalizer ($"Use `{Suite} help` for usage."));
+				return 1;
+			}
+			var command = Contains (extra [0]) ? this [extra [0]] : null;
+			if (command == null) {
+				help.WriteUnknownCommand (extra [0]);
+				return 1;
+			}
+			extra.RemoveAt (0);
+			if (showHelp) {
+				if (command.Options?.Contains ("help") ?? true) {
+					extra.Add ("--help");
+					return command.Invoke (extra);
+				}
+				command.Options.WriteOptionDescriptions (Out);
+				return 0;
+			}
+			return command.Invoke (extra);
+		}
+	}
+
+	public class HelpCommand : Command
+	{
+		public HelpCommand ()
+			: base ("help", help: "Show this message and exit")
+		{
+		}
+
+		public override int Invoke (IEnumerable<string> arguments)
+		{
+			var extra   = new List<string> (arguments ?? new string [0]);
+			var _       = CommandSet.Options.MessageLocalizer;
+			if (extra.Count == 0) {
+				CommandSet.Options.WriteOptionDescriptions (CommandSet.Out);
+				return 0;
+			}
+			var command = CommandSet.Contains (extra [0])
+				? CommandSet [extra [0]]
+				: null;
+			if (command == this || extra [0] == "--help") {
+				CommandSet.Out.WriteLine (_ ($"Usage: {CommandSet.Suite} COMMAND [OPTIONS]"));
+				CommandSet.Out.WriteLine (_ ($"Use `{CommandSet.Suite} help COMMAND` for help on a specific command."));
+				CommandSet.Out.WriteLine ();
+				CommandSet.Out.WriteLine (_ ($"Available commands:"));
+				CommandSet.Out.WriteLine ();
+				foreach (var c in CommandSet) {
+					CommandSet.Options.WriteCommandDescription (CommandSet.Out, c);
+				}
+				return 0;
+			}
+			if (command == null) {
+				WriteUnknownCommand (extra [0]);
+				return 1;
+			}
+			if (command.Options != null) {
+				command.Options.WriteOptionDescriptions (CommandSet.Out);
+				return 0;
+			}
+			return command.Invoke (new [] { "--help" });
+		}
+
+		internal void WriteUnknownCommand (string unknownCommand)
+		{
+			CommandSet.Error.WriteLine (CommandSet.Options.MessageLocalizer ($"{CommandSet.Suite}: Unknown command: {unknownCommand}"));
+			CommandSet.Error.WriteLine (CommandSet.Options.MessageLocalizer ($"{CommandSet.Suite}: Use `{CommandSet.Suite} help` for usage."));
 		}
 	}
 }
